/* 
   MPI Mandelbrot set computation

   Original program by Charles Calkins

   Modified by Jim Teresco, Williams College

   $Id: mand-mpi.cc,v 1.1 2003/03/04 05:28:42 terescoj Exp $
*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <iostream>

#include <mpi.h>
#include "mand.h"

using namespace std;

// only the master process does this - it sends out segments to
// compute to each slave process, then waits for each to finish up
double *DoMPI(int xsize, int ysize, double xmin, double xmax, 
	    double ymin, double ymax, int maxiter, double h, 
	    double thresh, int nprocs) {
  
  double *array=new double[xsize*ysize];
  if (!array) return NULL;
  
  int i;
  
  double delta=thresh*(xmax-xmin)/(xsize-1);
  
  int step;
  if (!nprocs)
    step=ysize;
  else
    step=ysize/(nprocs-1);
  
  // i is not 0 so you never send to the root (yourself)
  for (i=1; i<nprocs; i++) {
    int miniy=(i-1)*step;
    int maxiy=(i-1)*step+step-1;
    
    // send to process i
    MPI_Send(&miniy, 1, MPI_INT, i, 0, MPI_COMM_WORLD);
    MPI_Send(&maxiy, 1, MPI_INT, i, 0, MPI_COMM_WORLD);
    MPI_Send(&xmin, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD);
    MPI_Send(&xmax, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD);
    MPI_Send(&ymin, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD);
    MPI_Send(&ymax, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD);
    MPI_Send(&xsize, 1, MPI_INT, i, 0, MPI_COMM_WORLD);
    MPI_Send(&ysize, 1, MPI_INT, i, 0, MPI_COMM_WORLD);
    MPI_Send(&maxiter, 1, MPI_INT, i, 0, MPI_COMM_WORLD);
    MPI_Send(&h, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD);
    MPI_Send(&delta, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD);
  }

  // wait for processes to return array segments - there must be a better way

  for (i=1; i<nprocs; i++) {
    MPI_Status status;

    int miniy, maxiy;
    MPI_Recv(&miniy, 1, MPI_INT, i, 0, MPI_COMM_WORLD, &status);
    MPI_Recv(&maxiy, 1, MPI_INT, i, 0, MPI_COMM_WORLD, &status);

    // reassemble array segments into array  

    double *arraypart=new double[(maxiy-miniy+2)*xsize];
    MPI_Recv(arraypart, (maxiy-miniy+2)*xsize, MPI_DOUBLE, i, 0, 
	     MPI_COMM_WORLD, &status);
    for (int iy=miniy; iy<maxiy; iy++)
      for (int ix=0; ix<xsize; ix++)
	array[iy*xsize+ix]=arraypart[(iy-miniy)*xsize+ix];
    delete arraypart;
  }  
  return array;
}

// slave process, just waits for its instructions, then computes, then
// sends back answers
void DoSlave(int myid) {
  int miniy, maxiy, maxiter, xsize, ysize;
  double xmin,xmax, ymin,ymax,h, delta;
  MPI_Status status;
  
  // read parameters that are sent here
  MPI_Recv(&miniy, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
  MPI_Recv(&maxiy, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
  MPI_Recv(&xmin, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);
  MPI_Recv(&xmax, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);
  MPI_Recv(&ymin, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);
  MPI_Recv(&ymax, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);
  MPI_Recv(&xsize, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
  MPI_Recv(&ysize, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
  MPI_Recv(&maxiter, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
  MPI_Recv(&h, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);
  MPI_Recv(&delta, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);
  
  double dist,cx,cy;
  int iy;
  
  double *array=new double[(maxiy-miniy+2)*xsize];
  double *xorbit=new double[maxiter+1];
  double *yorbit=new double[maxiter+1];
  for (iy=miniy; iy<=maxiy; iy++) {
    cy=ymin+iy*(ymax-ymin)/(ysize-1);
    for (int ix=0; ix<xsize; ix++) {
      cx=xmin+ix*(xmax-xmin)/(xsize-1);
      dist=MSetDist(cx,cy,maxiter,h, xorbit, yorbit);
      if (dist<delta)
	array[(iy-miniy)*xsize+ix]=0;
      else
	array[(iy-miniy)*xsize+ix]=dist;		
    }
  }	
  
  delete xorbit;
  delete yorbit;
  
  // ship back the computed array (this piece of it), with iymin and 
  // iymax too so root knows what to reassemble
  
  MPI_Send(&miniy, 1, MPI_INT, 0, 0, MPI_COMM_WORLD);
  MPI_Send(&maxiy, 1, MPI_INT, 0, 0, MPI_COMM_WORLD);
  MPI_Send(array, (maxiy-miniy+2)*xsize, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
  delete array;
}

// this function is called on the master - it just parses the command
// line, prints parameters, and calls DoMPI, which sends out a subset
// of the work to each available slave process
void DoMaster(int argc, char *argv[], int nprocs) {
  double xmin, xmax, ymin, ymax, h, thresh;
  int maxiter, nprocs_ignore, xsize, ysize;
  double *array;
  char filename[256];
  
  parse_cmdline(argc, argv, 
		&xsize, &ysize,
		&xmin, &xmax, &ymin, &ymax, 
		&maxiter, &h, &thresh, &nprocs_ignore, filename);
  
  printf("xsize=%d  ysize=%d\n", xsize, ysize);
  printf("xmin=%f  xmax=%f  ymin=%f  ymax=%f\n",
	 xmin, xmax, ymin, ymax);
  printf("maxiter=%d  h=%f  thresh=%f  nprocs=%d\n", 
	 maxiter, h, thresh, nprocs);
  printf("filename=%s\n", filename);
  
  double start=MPI_Wtime();
  
  // ignore user nprocs, it's what the run command said that we have
  // to use
  array=DoMPI(xsize, ysize, xmin, xmax, ymin, ymax,
	    maxiter, h, thresh, nprocs);
  
  printf("Elapsed time: %f\n", (MPI_Wtime()-start)/1.0);
  
  if (array) {
    //SaveAsTGA(filename, xsize, ysize, array);
    delete array;
  }
  else 
    cerr << "Error in mpi" << endl;  
}

int main(int argc, char *argv[]) {
  int myid, nprocs;
  
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs); 
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);
  
  if (myid==0) 
    DoMaster(argc, argv, nprocs);
  else
    DoSlave(myid);
  
  MPI_Finalize();
  return 0;
}

